package POWER::Email;

use warnings;
use strict;
use Carp;

# update DEPENDENCIES in POD & Build.PL & README
use version; our $VERSION = qv('0.2.2');    # update POD & Changes & README

use Perl6::Export::Attrs;
use MIME::Base64;

use POWER::Utils;
use POWER::GPG;


my @Mail_Prog   = ('/usr/sbin/sendmail', '-t');
sub _send_mail {
    my ($email) = @_;
    pipe my $r_fh, my $w_fh or die "Email: pipe: $!";
    my ($exitcode, $signal) = System({
        cmd     => \@Mail_Prog,
        fh      => [ $r_fh ],
        timeout => 30,
        cb      => sub {
            close $r_fh;
            print {$w_fh} $email;
            close $w_fh;
        }
    });
    die "Email: sendmail was killed by signal $signal\n"    if $signal;
    die "Email: sendmail died with exit code $exitcode\n"   if $exitcode;
    return;
}

#    :
#   RFC2822, 2.1
#          CRLF.
#   DJB
#        "few broken gateway"      
#       .     ( RFC822,  
#       RFC2822) " " (()  CRLF - folding).
#         ģ     :
#              -   
#                .
#   me
#        ,   Σ     "few broken
#       gateway"...       ,   
#           qmail   .
#          CRLF       LF.
# ($head, $body) = _split2headbody($email);
# Return:    :   .
#          ,   .
#      -   .
#        (CRLF  LF).
# Throws:         .
sub _split2headbody {
    return ($1, $2) if $_[0] =~ /(.*?\n)\r?\n(.*)/s;
    croak "no head/body separator found";
}

# @headers = %{ _get_headers($header) };
# Parse headers according to RFC 822:
# - header unfolding
# - strip comments
# Return array with even elements - header names in lower case and odd
# elements - header values.
my $comment822; 
$comment822 = qr/\( (?:  [^()\\\n]  |  \\.  |  (??{ $comment822 })  )* \)/xms;
my $quoted822   = qr/" (?:  [^"\\\n]  |  \\.  )* "/xms;
sub _get_headers {
    my ($head) = @_;
    $head =~ s/[ \t]*\n[ \t]+/ /g; # Unfolding
    my @res;
    while ($head =~ /^([^:\s]+):[ \t]*(.*)$/gm) {
        my ($h, $c) = (lc($1), $2);
        $c =~ s/\G  ((?:  [^("]  |  $quoted822  )*)  $comment822/$1 /xmsg;
        push @res, $h => $c;
    }
    return @res;
}

# Email ($tmpl, \%data, \%attach, \%gpg);
# WARNING   Email() unable to handle complex (MIME) templates if called with
#           attach/gpg params.
# Email($tmpl, \%data, \%attach, {
#   encr_to     => \@gpgids || [],
#   sign_by     => \@gpgids || [],
# });
sub Email :Export(:DEFAULT) {
    my ($tmpl, $hash_ref, $attach, $gpg) = @_;
    $tmpl = Cat($tmpl) if $tmpl !~ /\r|\n/;
    my $s = q{};
    while (1) {
        if    ($tmpl =~ /\G@~(.*?)~@/gcs) {
            $s .= !length $1                ? croak "wrong template: @~~@"
                : defined $hash_ref->{$1}   ? $hash_ref->{$1}
                :                             croak "no value for @~$1~@"
                ;
        }
        elsif ($tmpl =~ /\G(@~|~@)/gcs) {
            croak "wrong template: orphaned $1";
        }
        elsif ($tmpl =~ /\G(.+?)(?=@~|~@|\z)/gcs) {
            $s .= $1;
        }
        else {
            last;
        }
    }
    # Extract general headers and body from template.
    my ($head, $body) = _split2headbody($s);
    $body = q{} if !defined $body;
    # Assure $head and $body end with \n (to add Content-Type to $head or
    # boundaries to $body safely, if needed).
    s/\n?\z/\n/ for $head, $body;
    # Extract content-related headers from general headers.
    # At this point $chead will contain content headers for email text.
    my $chead = q{};
    while ($head =~ s/^(Content-.*\n(?:[ \t].*\n)*)//mi) {
        $chead .= $1;
    }
    if ($chead !~ /^Content-Type:/mi) {
        $chead .= "Content-Type: TEXT/PLAIN; charset=US-ASCII\n"
    }
    # If there files to attach, convert email into MIME.
    if ($attach) {
	# Update content-headers and body according MULTIPART/MIXED.
        my $bound = RandCh(35);
        $body = "--$bound\n".$chead."\n".$body;
        for my $file (keys %$attach) { # Prepare attachment
            $attach->{$file} = Cat($file) if !defined $attach->{$file};
            (my $filename = $file) =~ s!.*/!!s;
            $filename =~ s/\s/ /g;
            $filename =~ s/(["\\])/\\$1/g;
            $body 
               .= "--$bound\n"
                . "Content-Type: APPLICATION/octet-stream; name=\"$filename\"\n"
                . "Content-Transfer-Encoding: BASE64\n"
                . "Content-Disposition: attachment; filename=\"$filename\"\n"
                . "\n"
                . encode_base64($attach->{$file})
                ;
        }
        $body .= "--$bound--\n";
        # From this point $chead will contain content headers for MIME
        # with attached files.
        $chead
            = "MIME-Version: 1.0\n"
            . "Content-Type: MULTIPART/MIXED; BOUNDARY=\"$bound\"\n"
            ;
    }
    if ($gpg) {
        croak "require at least one of {encr_to} or {sign_by} keys."
            if !exists $gpg->{encr_to} && !exists $gpg->{sign_by};
        defined $gpg->{$_} or croak "{$_} undefined"
            for grep {exists $gpg->{$_}} qw( encr_to sign_by );
        my $bound = RandCh(40);
        my $gpg_body = $chead."\n".$body;
        # For encrypt or encrypt/sign case
        if (defined $gpg->{encr_to}) {
            if (!@{ $gpg->{encr_to} }) {
                my @headers = _get_headers($head);
                while (@headers) {
                    my ($h, $c) = splice @headers, 0, 2;
                    next if $h !~ /\A (?: to | cc | bcc ) \z/xms;
                    # FIXME non RFC822-correct:
                    for ($c =~ m/\G((?:[^",\\]|\\.|$quoted822)+),?/g) {
                        s/\A\s+//;
                        s/\s+\z//;
                        s/.*<([^<>]+)>\z/$1/;
                        push @{ $gpg->{encr_to} }, $_;
                    }
                }
            }
            my $res = $gpg->{sign_by}
                ? gpg_encrypt_sign($gpg_body, $gpg->{encr_to}, $gpg->{sign_by})
                :      gpg_encrypt($gpg_body, $gpg->{encr_to})
                ;
            # Update content-headers and body according MULTIPART/ENCRYPTED.
            $body =
                "--$bound\n".
                "Content-Type: application/pgp-encrypted\n".
                "Content-Disposition: attachment\n".
                "\n".
                "Version: 1\n".
                "\n".
                "--$bound\n".
                "Content-Type: application/octet-stream\n".
                "Content-Disposition: inline; filename=\"msg.asc\"\n".
                "\n".
                $res.
                "\n".
                "--$bound--\n";
            $chead =
                "MIME-Version: 1.0\n".
                "Content-Type: MULTIPART/ENCRYPTED; protocol=\"application/pgp-encrypted\"; BOUNDARY=\"$bound\"\n".
                "Content-Disposition: inline\n";
        }
        # For only sign case with attach
        elsif (defined $gpg->{sign_by}) {
            # Update content-headers and body according MULTIPART/SIGNED.
            my $res = gpg_detachsign($gpg_body, $gpg->{sign_by});
            $body =
                "--$bound\n".
                $gpg_body."\n".
                "--$bound\n".
                "Content-Type: application/pgp-signature\n".
                "Content-Disposition: inline\n".
                "\n".
                $res.
                "\n".
                "--$bound--\n";
            $chead =
                "MIME-Version: 1.0\n".
                "Content-Type: MULTIPART/SIGNED; micalg=\"pgp-sha1\"; protocol=\"application/pgp-signature\"; BOUNDARY=\"$bound\"\n".
                "Content-Disposition: inline\n";
        }
    }
    _send_mail($head.$chead."\n".$body);
    return;
}





1; # Magic true value required at end of module
__END__

=head1 NAME

POWER::Email - email handling.


=head1 VERSION

This document describes POWER::Email version 0.2.2


=head1 SYNOPSIS

    use POWER::Email;

    Email($template, \%fields);


=head1 DESCRIPTION



=head1 EXPORTS

=over

=item :DEFAULT

Email()


=back


=head1 INTERFACE 

=over

=item B<Email>( $template, $data [, $attach [, $gpg]])

Simple (and non RFC compliant) email sending. Features:

 - Process simple templates (just substitute scalar values into
   placeholders, no conditional/loop/code execution supported).
   - Template can be read from disk or used from scalar.
   - Checking template syntax.
   - Checking for undefined values for placeholders.
   - Automatically add Content-Type field if template doesn't
     contain it.
 - Can convert simple non-MIME templates into MIME to attach
   files and/or encrypt/sign with GPG. (Attaching files or
   encrypt/sign with GPG for MIME templates not supported!)
 - Attached files can be read from disk or scalar.
 - Recipient list for GPG encryption can be detected automatically
   from To/Cc/Bcc fields. (WARNING: this feature non RFC compliant
   so complex emails can be detected incorrectly!)
 - Mail injected using `/usr/bin/sendmail -t`.

If $template looks like a file name (doesn't contain newline)
then email template will be read from it.
In other case scalar $template will be treated as template text.

If template doesn't contain Content-Type field this one will be added:

 Content-Type: TEXT/PLAIN; charset=US-ASCII

$data must be a hash reference, and this hash will be used to substitute
placeholders in email template.
%$data MUST contain defined values for all placeholders.
Placeholders format is: @~field_name~@ .

$attach must be a hash reference. Keys in this hash treated as file names.
If %$attach element has defined value this value used as file content, else
file content is read from disk.

If $attach is set, then template will be automatically reformatted to
MIME MULTIPART/MIXED format. First part will be template body, second and other
parts will be attached files. All /^Content-/ headers found in template
will be moved to first part's headers.

$gpg must be a hash reference.
This hash should contain at least one of keys 'sign_by' or 'encr_to'
(or both):

 sign_by    SCALAR sender key identification string for gpg
 encr_to    ARRAYREF with recipients identification strings for gpg

If {encr_to} is empty array then recipient list will be automatically
detected from To/Cc/Bcc fields. WARNING: this may reveal list of Bcc
recipients to To/Cc recipients!

If $gpg is set, then template will be automatically reformatted to
MIME MULTIPART/ENCRYPTED or MULTIPART/SIGNED formats.


Example:

 $template = <<EOT;
 From: Email robot <nobody>
 To: @~email~@
 Subject: @~subj~@
 Content-Type: text/plain; charset=Windows-1251
 Content-Transfer-Encoding: 8bit

 Dear @~name~@.
 This message is a test.
 
 Sincerely, Robot.
 EOT
 $data = {
    email => 'user@remote.host.com',
    subj  => "Test",
    name  => "John Doe",
    };
 $attach = {
    "outbound/pic/1.jpg" => undef,
    "comment.txt" => "1.jpg is my photo.\n\nJonh.",
    };
 Email($template, $data, $attach);


=back


=head1 DIAGNOSTICS

=over

=item C<< Email: $! >>

Email(): try to open pipe to '/usr/sbin/sendmail -t' and fail.


=item C<< no head/body separator found >>

Email(): template doesn't contain empty line after header.


=item C<< wrong template: @~~@ >>

=item C<< wrong template: orphaned @~ >>

=item C<< wrong template: orphaned ~@ >>

Email(): syntax error in template.


=item C<< no value for @~...~@ >>

Email(): template has placeholder, but %$data doesn't define value for it.


=item C<< require at least one of {encr_to} or {sign_by} keys. >>

=item C<< {encr_to} undefined >>

=item C<< {sign_by} undefined >>

Email(): %$gpg must contain at least one of these keys and their values
must be defined.


=back


=head1 CONFIGURATION AND ENVIRONMENT

POWER::Email requires no configuration files or environment variables.


=head1 DEPENDENCIES

=over

=item version

=item Perl6::Export::Attrs

=item POWER::Utils

=item POWER::GPG


=back


=head1 INCOMPATIBILITIES

None reported.


=head1 BUGS AND LIMITATIONS

No bugs have been reported.

Please report any bugs or feature requests to
C<bug-power-email@rt.cpan.org>, or through the web interface at
L<http://rt.cpan.org>.


=head1 AUTHOR

Alex Efros  C<< <powerman@asdfgroup.com> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2005,2006, Alex Efros C<< <powerman@asdfgroup.com> >>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
